#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#
# The InverseProblem class is based on the SingleOutputProblem and
# MultiOutputProblem classes of PINTS (https://github.com/pints-team/pints/),
# which is distributed under the BSD 3-clause license.
#

import copy
from warnings import warn

import myokit
import numpy as np
import pandas as pd
import pints

import chi


class ProblemModellingController(object):
    """
    A problem modelling controller which simplifies the model building process
    of a pharmacokinetic and pharmacodynamic problem.

    The class is instantiated with an instance of a :class:`MechanisticModel`
    and one instance of an :class:`ErrorModel` for each mechanistic model
    output.

    :param mechanistic_model: A mechanistic model for the problem.
    :type mechanistic_model: MechanisticModel
    :param error_models: A list of error models. One error model has to be
        provided for each mechanistic model output.
    :type error_models: list[ErrorModel]
    :param outputs: A list of mechanistic model output names, which can be used
        to map the error models to mechanistic model outputs. If ``None``, the
        error models are assumed to be ordered in the same order as
        :meth:`MechanisticModel.outputs`.
    :type outputs: list[str], optional
    """

    def __init__(self, mechanistic_model, error_models, outputs=None):
        super(ProblemModellingController, self).__init__()

        # Check inputs
        if not isinstance(mechanistic_model, chi.MechanisticModel):
            raise TypeError(
                'The mechanistic model has to be an instance of a '
                'chi.MechanisticModel.')

        if not isinstance(error_models, list):
            error_models = [error_models]

        for error_model in error_models:
            if not isinstance(error_model, chi.ErrorModel):
                raise TypeError(
                    'Error models have to be instances of a '
                    'chi.ErrorModel.')

        # Copy mechanistic model
        mechanistic_model = mechanistic_model.copy()

        # Set outputs
        if outputs is not None:
            mechanistic_model.set_outputs(outputs)

        # Get number of outputs
        n_outputs = mechanistic_model.n_outputs()

        if len(error_models) != n_outputs:
            raise ValueError(
                'Wrong number of error models. One error model has to be '
                'provided for each mechanistic error model.')

        # Copy error models
        error_models = [copy.copy(error_model) for error_model in error_models]

        # Remember models
        self._mechanistic_model = mechanistic_model
        self._error_models = error_models

        # Set defaults
        self._population_model = None
        self._log_prior = None
        self._data = None
        self._covariates = None
        self._dosing_regimens = None

        self._set_error_model_parameter_names()

    def _check_covariate_dict(
            self, covariate_dict, covariate_names, observables):
        """
        Makes sure that for each covariate name (from model) an observable
        (from dataframe) exists.
        """
        # Check that model needs covariates
        if len(covariate_names) == 0:
            return None

        # If no mapping is provided, construct default mapping
        if covariate_dict is None:
            # Assume trivial map
            covariate_dict = {cov: cov for cov in covariate_names}

        # Check that covariate name map is valid
        for cov in covariate_names:
            if cov not in list(covariate_dict.keys()):
                raise ValueError(
                    'The covariate <' + str(cov) + '> could not be identified '
                    'in the covariate name map.')

            mapped_name = covariate_dict[cov]
            if mapped_name not in observables:
                raise ValueError(
                    'The covariate <' + str(mapped_name) + '> could not be '
                    'identified in the dataframe.')

        return covariate_dict

    def _check_covariate_values(self, covariate_names):
        """
        Makes sure that covariates can be reshaped in to an array of shape
        (n, c).

        In other words, checks whether for each covariate_name there exists
        exactly one non-NaN value for each ID.
        """
        # Check that model needs covariates
        if len(covariate_names) == 0:
            return None

        for name in covariate_names:
            # Mask covariate values
            mask = self._data[self._obs_key] == self._covariate_dict[name]
            temp = self._data[mask]

            for _id in self._ids:
                # Mask values for individual
                mask = temp[self._id_key] == _id
                temp2 = temp.loc[mask, self._value_key].dropna()

                if len(temp2) != 1:
                    covariate = self._covariate_dict[name]
                    raise ValueError(
                        'There are either 0 or more than 1 value of the '
                        'covariate %s for ID %s. '
                        'Exactly one covariate value '
                        'has to be provided for each ID.' % (covariate, _id)
                    )

    def _check_output_observable_dict(
            self, output_observable_dict, outputs, observables):
        """
        Makes sure that the mechanistic model outputs are correctly mapped to
        the observables in the dataframe.
        """
        if output_observable_dict is None:
            if (len(outputs) == 1) and (len(observables) == 1):
                # Create map of single output to single observable
                output_observable_dict = {outputs[0]: observables[0]}
            else:
                # Assume trivial map
                output_observable_dict = {output: output for output in outputs}

        # Check that output-observable map is valid
        for output in outputs:
            if output not in list(output_observable_dict.keys()):
                raise ValueError(
                    'The output <' + str(output) + '> could not be identified '
                    'in the output-observable map.')

            observable = output_observable_dict[output]
            if observable not in observables:
                raise ValueError(
                    'The observable <' + str(observable) + '> could not be '
                    'identified in the dataframe.')

        return output_observable_dict

    def _clean_data(self, dose_key, dose_duration_key):
        """
        Makes sure that the data is formated properly.

        1. ids are strings
        2. time are numerics or NaN
        3. observables are strings
        4. values are numerics or NaN
        5. observable types are 'Modelled' or 'Covariate'
        6. dose are numerics or NaN
        7. duration are numerics or NaN
        """
        # Create container for data
        columns = [
            self._id_key, self._time_key, self._obs_key, self._value_key]
        if dose_key is not None:
            columns += [dose_key]
        if dose_duration_key is not None:
            columns += [dose_duration_key]
        data = pd.DataFrame(columns=columns)

        # Convert IDs to strings
        data[self._id_key] = self._data[self._id_key].astype(
            "string")

        # Convert times to numerics
        data[self._time_key] = pd.to_numeric(self._data[self._time_key])

        # Convert observables to strings
        data[self._obs_key] = self._data[self._obs_key].astype(
            "string")

        # Convert values to numerics
        data[self._value_key] = pd.to_numeric(self._data[self._value_key])

        # Convert dose to numerics
        if dose_key is not None:
            data[dose_key] = pd.to_numeric(
                self._data[dose_key])

        # Convert duration to numerics
        if dose_duration_key is not None:
            data[dose_duration_key] = pd.to_numeric(
                self._data[dose_duration_key])

        self._data = data

    def _create_hierarchical_log_likelihood(self, log_likelihoods):
        """
        Returns an instance of a chi.HierarchicalLoglikelihood based on
        the provided list of log-likelihoods and the population models.
        """
        # Get covariates from the dataset if any are needed
        covariate_names = self.get_covariate_names()
        covariates = None
        if len(covariate_names) > 0:
            covariates = self._extract_covariates(covariate_names)

        log_likelihood = chi.HierarchicalLogLikelihood(
                log_likelihoods, self._population_model, covariates)

        return log_likelihood

    def _create_log_likelihoods(self, ids):
        """
        Returns a list of log-likelihoods, one for each individual in the
        dataset. If individual is not None, only the likelihood of this
        individual is returned.
        """
        # Create a likelihood for each individual
        log_likelihoods = []
        for individual in ids:
            # Set dosing regimen
            if self._dosing_regimens:
                self._mechanistic_model.set_dosing_regimen(
                    self._dosing_regimens[individual])

            log_likelihood = self._create_log_likelihood(individual)
            if log_likelihood is not None:
                # If data exists for this individual, append to log-likelihoods
                log_likelihoods.append(log_likelihood)

        if len(ids) == 1:
            return log_likelihoods[0]

        return log_likelihoods

    def _create_log_likelihood(self, individual):
        """
        Gets the relevant data for the individual and returns the resulting
        chi.LogLikelihood.
        """
        # Get individuals data
        times = []
        observations = []
        mask = self._data[self._id_key] == individual
        data = self._data[mask][
            [self._time_key, self._obs_key, self._value_key]]
        for output in self._mechanistic_model.outputs():
            # Mask data for observable
            observable = self._output_observable_dict[output]
            mask = data[self._obs_key] == observable
            temp_df = data[mask]

            # Filter times and observations for non-NaN entries
            mask = temp_df[self._value_key].notnull()
            temp_df = temp_df[[self._time_key, self._value_key]][mask]
            mask = temp_df[self._time_key].notnull()
            temp_df = temp_df[mask]

            # Collect data for output
            times.append(temp_df[self._time_key].to_numpy())
            observations.append(temp_df[self._value_key].to_numpy())

        # # Count outputs that were measured
        # # TODO: copy mechanistic model and update model outputs.
        # # (Useful for e.g. control group and dose group training)
        # n_measured_outputs = 0
        # for output_measurements in observations:
        #     if len(output_measurements) > 0:
        #         n_measured_outputs += 1

        # Create log-likelihood and set ID to individual
        log_likelihood = chi.LogLikelihood(
            self._mechanistic_model, self._error_models, observations, times)
        log_likelihood.set_id(individual)

        return log_likelihood

    def _extract_covariates(self, covariate_names):
        """
        Extracts covariates from the pandas.DataFrame and formats them
        as a np.ndarray of shape (n, c).
        """
        # Format covariates to array of shape (n, c)
        c = len(covariate_names)
        n = len(self._ids)
        covariates = np.empty(shape=(n, c))
        for idc, name in enumerate(covariate_names):
            mask = self._data[self._obs_key] == self._covariate_dict[name]
            temp = self._data[mask]
            for idn, _id in enumerate(self._ids):
                mask = temp[self._id_key] == _id
                covariates[idn, idc] = \
                    temp.loc[mask, self._value_key].dropna().values

        return covariates

    def _extract_dosing_regimens(self, dose_key, duration_key):
        """
        Converts the dosing regimens defined by the pandas.DataFrame into
        myokit.Protocols, and returns them as a dictionary with individual
        IDs as keys, and regimens as values.

        For each dose entry in the dataframe a dose event is added
        to the myokit.Protocol. If the duration of the dose is not provided
        a bolus dose of duration 0.01 time units is assumed.
        """
        # Create duration column if it doesn't exist and set it to default
        # bolus duration of 0.01
        if duration_key is None:
            duration_key = 'Duration in base time unit'
            self._data[duration_key] = 0.01

        # Extract regimen from dataset
        regimens = dict()
        for label in self._ids:
            # Filter times and dose events for non-NaN entries
            mask = self._data[self._id_key] == label
            data = self._data[
                [self._time_key, dose_key, duration_key]][mask]
            mask = data[dose_key].notnull()
            data = data[mask]
            mask = data[self._time_key].notnull()
            data = data[mask]

            # Add dose events to dosing regimen
            regimen = myokit.Protocol()
            for _, row in data.iterrows():
                # Set duration
                duration = row[duration_key]
                if np.isnan(duration):
                    # If duration is not provided, we assume a bolus dose
                    # which we approximate by 0.01 time_units.
                    duration = 0.01

                # Compute dose rate and set regimen
                dose_rate = row[dose_key] / duration
                time = row[self._time_key]
                regimen.add(myokit.ProtocolEvent(dose_rate, time, duration))

            regimens[label] = regimen

        return regimens

    def _set_error_model_parameter_names(self):
        """
        Resets the error model parameter names and prepends the output name
        if more than one output exists.
        """
        # Reset error model parameter names to defaults
        for error_model in self._error_models:
            error_model.set_parameter_names(None)

        # Rename error model parameters, if more than one output
        n_outputs = self._mechanistic_model.n_outputs()
        if n_outputs > 1:
            # Get output names
            outputs = self._mechanistic_model.outputs()

            for output_id, error_model in enumerate(self._error_models):
                # Get original parameter names
                names = error_model.get_parameter_names()

                # Prepend output name
                output = outputs[output_id]
                names = [output + ' ' + name for name in names]

                # Set new parameter names
                error_model.set_parameter_names(names)

    def fix_parameters(self, name_value_dict):
        """
        Fixes the value of model parameters, and effectively removes them as a
        parameter from the model. Fixing the value of a parameter to ``None``,
        sets the parameter free again.

        .. note::
            1. Fixing model parameters resets the log-prior to ``None``.
            2. Once a population model is set, only population model
               parameters can be fixed.
            3. Setting data resets all population parameters as the number of
               parameters may change with the number of modelled individuals.

        :param name_value_dict: A dictionary with model parameters as keys, and
            the value to be fixed at as values.
        :type name_value_dict: dict
        """
        # Check type of dictionanry
        try:
            name_value_dict = dict(name_value_dict)
        except (TypeError, ValueError):
            raise ValueError(
                'The name-value dictionary has to be convertable to a python '
                'dictionary.')

        # Convert models to reduced models
        models = []
        if self._population_model is not None:
            # If a population model is set, fix only population parameters
            population_model = self._population_model
            if not isinstance(population_model, chi.ReducedPopulationModel):
                population_model = chi.ReducedPopulationModel(
                    population_model)
            models.append(population_model)
        else:
            mechanistic_model = self._mechanistic_model
            error_models = self._error_models
            if not isinstance(mechanistic_model, chi.ReducedMechanisticModel):
                mechanistic_model = chi.ReducedMechanisticModel(
                    mechanistic_model)
            for idm, error_model in enumerate(error_models):
                if not isinstance(error_model, chi.ReducedErrorModel):
                    error_model = chi.ReducedErrorModel(error_model)
                error_models[idm] = error_model
            models += [mechanistic_model] + error_models

        # Fix parameters
        for model in models:
            model.fix_parameters(name_value_dict)

        # Safe models
        # (if no parameters are fixed, convert back to non-reduced models)
        if self._population_model is not None:
            population_model = models[0]
            if population_model.n_fixed_parameters() == 0:
                population_model = population_model.get_population_model()
            self._population_model = population_model
        else:
            mechanistic_model = models[0]
            error_models = models[1:]
            if mechanistic_model.n_fixed_parameters() == 0:
                mechanistic_model = mechanistic_model.mechanistic_model()
            for idm, error_model in enumerate(error_models):
                if error_model.n_fixed_parameters() == 0:
                    error_models[idm] = error_model.get_error_model()
            self._mechanistic_model = mechanistic_model
            self._error_models = error_models

        # Reset priors
        self._log_prior = None

    def get_dosing_regimens(self):
        """
        Returns a dictionary of dosing regimens in form of
        :class:`myokit.Protocol` instances.

        The dosing regimens are extracted from the dataset if a dose key is
        provided. If no dose key is provided ``None`` is returned.
        """
        return self._dosing_regimens

    def get_log_prior(self):
        """
        Returns the :class:`LogPrior` for the model parameters. If no
        log-prior is set, ``None`` is returned.
        """
        return self._log_prior

    def get_log_posterior(self, individual=None):
        r"""
        Returns the :class:`LogPosterior` defined by the measurements of the
        modelled observables,
        the administered dosing regimen, the mechanistic model, the error
        model, the log-prior, and optionally the population model, covariates
        and the fixed model parameters.

        If measurements of multiple individuals exist in the dataset, the
        indiviudals ID can be passed to return the log-posterior associated
        to that individual. If no ID is selected and no population model
        has been set, a list of log-posteriors is returned corresponding to
        each of the individuals.

        This method raises an error if the data or the log-prior have not been
        set. See :meth:`set_data` and :meth:`set_log_prior`.

        :param individual: The ID of an individual. If ``None`` the
            log-posteriors for all individuals is returned. Input is ignored if
            a population model is set.
        :type individual: str, optional
        """
        # Check prerequesites
        if self._data is None:
            raise ValueError(
                'The data has not been set.')
        if self._log_prior is None:
            raise ValueError(
                'The log-prior has not been set.')

        # Check that individual is in ids
        _id = individual if self._population_model is None else None
        if (_id is not None) and (_id not in self._ids):
            raise ValueError(
                'The individual cannot be found in the ID column of the '
                'dataset.')

        # Set ID to all IDs if population model is set, and to first ID
        # if not set but None
        if self._population_model is not None:
            ids = self._ids
        elif _id is None:
            ids = [self._ids[0]]
        else:
            ids = [_id]

        # Create log-likelihoods
        log_likelihood = self._create_log_likelihoods(ids)
        if self._population_model is not None:
            # Compose HierarchicalLogLikelihoods
            log_likelihood = self._create_hierarchical_log_likelihood(
                log_likelihood)

        # Compose the log-posterior
        if isinstance(log_likelihood, chi.LogLikelihood):
            log_posterior = chi.LogPosterior(
                log_likelihood, self._log_prior)
        else:
            log_posterior = chi.HierarchicalLogPosterior(
                log_likelihood, self._log_prior)

        return log_posterior

    def get_n_parameters(self, exclude_pop_model=False):
        """
        Returns the number of the model parameters, i.e. the combined number of
        parameters from the mechanistic model and the error model when no
        population model has been set, or the number of population model
        parameters when a population model has been set.

        Any parameters that have been fixed to a constant value will not be
        included in the number of model parameters.

        :param exclude_pop_model: A boolean flag to indicate whether the
            population model should be ignored (if set).
        :type exclude_pop_model: bool, optional
        """
        if (self._population_model is None) or exclude_pop_model:
            n_parameters = self._mechanistic_model.n_parameters()
            for error_model in self._error_models:
                n_parameters += error_model.n_parameters()
            return n_parameters

        return self._population_model.n_parameters()

    def get_covariate_names(self):
        """
        Returns the names of the covariates.
        """
        if self._population_model is None:
            return []

        return self._population_model.get_covariate_names()

    def get_parameter_names(self, exclude_pop_model=False):
        """
        Returns the names of the model parameters, i.e. the combined names of
        the mechanistic model parameters and the error model parameters when no
        population model has been set, or the names of population model
        parameters when a population model has been set.

        Any parameters that have been fixed to a constant value will not be
        included.

        :param exclude_pop_model: A boolean flag to indicate whether the
            population model should be ignored (if set).
        :type exclude_pop_model: bool, optional
        """
        if (self._population_model is None) or exclude_pop_model:
            names = self._mechanistic_model.parameters()
            for error_model in self._error_models:
                names += error_model.get_parameter_names()
            return names

        return self._population_model.get_parameter_names()

    def get_predictive_model(self):
        """
        Returns a :class:`PredictiveModel` defined by the mechanistic model,
        the error model, and optionally the population model and the
        fixed model parameters.
        """
        # Create predictive model
        predictive_model = chi.PredictiveModel(
            self._mechanistic_model, self._error_models)
        if (self._population_model is not None):
            predictive_model = chi.PopulationPredictiveModel(
                predictive_model, self._population_model)

        return predictive_model

    def set_data(
            self, data, output_observable_dict=None, covariate_dict=None,
            id_key='ID', time_key='Time', obs_key='Observable',
            value_key='Value', dose_key='Dose', dose_duration_key='Duration'):
        """
        Sets the data of the modelling problem.

        The data contains information about the measurement time points, the
        measured values of the observables, the observable name, IDs to
        identify the corresponding individuals, and optionally information
        on the administered dose amount and duration.

        The data is expected to be in form of a :class:`pandas.DataFrame`
        with the columns ID | Time | Observable | Value | Dose | Duration.

        If no information exists, the corresponding column
        keys can be set to ``None``.

        .. note::
            Time-dependent covariates are currently not supported. Thus, the
            Time column of observables that are used as covariates is ignored.

        :param data: A dataframe with an ID, time, observable,
            value and optionally a dose and duration column.
        :type data: pandas.DataFrame
        :param output_observable_dict: A dictionary with mechanistic model
            output names as keys and dataframe observable names as values. If
            ``None`` the model outputs and observables are assumed to have the
            same names.
        :type output_observable_dict: dict, optional
        :param covariate_dict: A dictionary with population model covariate
            names as keys and dataframe observables as values. If
            ``None`` the model covariates and observables are assumed to have
            the same names.
        :type covariate_dict: dict, optional
        :param id_key: The key of the ID column in the
            :class:`pandas.DataFrame`. Default is ``'ID'``.
        :type id_key: str, optional
        :param time_key: The key of the time column in the
            :class:`pandas.DataFrame`. Default is ``'Time'``.
        :type time_key: str, optional
        :param obs_key: The key of the observable column in the
            :class:`pandas.DataFrame`. Default is ``'Observable'``.
        :type obs_key: str, optional
        :param value_key: The key of the value column in the
            :class:`pandas.DataFrame`. Default is ``'Value'``.
        :type value_key: str, optional
        :param dose_key: The key of the dose column in the
            :class:`pandas.DataFrame`. Default is ``'Dose'``.
        :type dose_key: str, optional
        :param dose_duration_key: The key of the duration column in the
            :class:`pandas.DataFrame`. Default is ``'Duration'``.
        :type dose_duration_key: str, optional
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be a pandas.DataFrame.')

        # If model does not support dose administration, set dose keys to None
        if not self._mechanistic_model.supports_dosing():
            dose_key = None
            dose_duration_key = None

        keys = [id_key, time_key, obs_key, value_key]
        if dose_key is not None:
            keys += [dose_key]
        if dose_duration_key is not None:
            keys += [dose_duration_key]

        for key in keys:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')

        # Check output observable map
        outputs = self._mechanistic_model.outputs()
        observables = data[obs_key].dropna().unique()
        output_observable_dict = self._check_output_observable_dict(
            output_observable_dict, outputs, observables)

        # Check covariate name map
        covariate_names = self.get_covariate_names()
        covariate_dict = self._check_covariate_dict(
            covariate_dict, covariate_names, observables)

        self._id_key, self._time_key, self._obs_key, self._value_key = [
            id_key, time_key, obs_key, value_key]
        self._data = data[keys]
        self._output_observable_dict = output_observable_dict
        self._covariate_dict = covariate_dict

        # Make sure data is formatted correctly
        self._clean_data(dose_key, dose_duration_key)
        self._ids = self._data[self._id_key].unique()

        # Extract dosing regimens
        self._dosing_regimens = None
        if dose_key is not None:
            self._dosing_regimens = self._extract_dosing_regimens(
                dose_key, dose_duration_key)

        # Set number of modelled individuals of population model
        if isinstance(self._population_model, chi.ReducedPopulationModel):
            # Unfix model parameters
            self._population_model = \
                self._population_model.get_population_model()
        if self._population_model is not None:
            self._population_model.set_n_ids(len(self._ids))

        # Check that covariates can be reshaped into (n, c)
        self._check_covariate_values(covariate_names)

    def set_log_prior(self, log_prior):
        """
        Sets the prior distribution of the model parameters.

        The log-prior dimensions are assumed to be ordered according to
        :meth:`get_parameter_names`.

        .. note::
            This method requires that the data has been set, since the number
            of parameters of an hierarchical model may vary with the number
            of individuals in the dataset
            (see e.g. :class:`HeterogeneousModel`).

        :param log_prior: A :class:`pints.LogPrior` of the length
            :meth:`get_n_parameters`.
        :type log_priors: pints.LogPrior
        """
        # Check prerequesites
        if self._data is None:
            raise ValueError('The data has not been set.')

        # Check inputs
        if not isinstance(log_prior, pints.LogPrior):
            raise ValueError(
                'The log-prior has to be an instance of pints.LogPrior.')
        if log_prior.n_parameters() != self.get_n_parameters():
            raise ValueError(
                'The dimension of the log-prior has to be the same as the '
                'number of parameters in the model. There are <'
                + str(self.get_n_parameters()) + '> model parameters.')

        self._log_prior = log_prior

    def set_population_model(self, population_model):
        """
        Sets the population model.

        A population model specifies how model parameters vary across
        individuals. The dimension of the population model has to match the
        number of model parameters.

        .. note::
            Setting a population model resets the log-prior to ``None``,
            because it changes the top-level parameters of the model.

        :param population_model: A :class:`PopulationModel` whose dimension is
            the same as the number of bottom-level parameters.
        :type population_model: PopulationModel
        """
        # Check inputs
        if not isinstance(population_model, chi.PopulationModel):
            raise TypeError(
                'The population model has to be an instance of '
                'chi.PopulationModel.')

        # Make sure that dimension of population model is correct
        n_parameters = self.get_n_parameters(exclude_pop_model=True)
        if population_model.n_dim() != n_parameters:
            raise ValueError(
                'The dimension of the population model does not match the '
                'number of bottom-level parameters. There are '
                '<' + str(n_parameters) + '> bottom-level parameters.')

        # Remember population model
        self._population_model = population_model

        # Set number of modelled individuals, if data has been set already
        if self._data is not None:
            self._population_model.set_n_ids(len(self._ids))

        # Set dimension names to bottom-level parameters
        names = self.get_parameter_names(exclude_pop_model=True)
        self._population_model.set_dim_names(names)

        # Check that covariates can be found, if data has already been set
        if self._data is not None:
            try:
                # Get covariate names
                covariate_names = self.get_covariate_names()
                observables = self._data[self._obs_key].dropna().unique()
                self._covariate_dict = self._check_covariate_dict(
                    self._covariate_dict, covariate_names, observables)
            except ValueError:
                # New population model and data are not compatible, so reset
                # data
                self._data = None
                warn(
                    'The covariates of the new population model could not '
                    'automatically be matched to the observables in the '
                    'dataset. The data was therefore reset. Please set the '
                    'data again with the `set_data` method and specify the '
                    'covariate mapping.', UserWarning)

        # Set prior to default
        self._log_prior = None



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import numpy as np


class PopulationFilter(object):
    r"""
    A base class for filters.

    A filter estimates the likelihood with which simulated
    observations :math:`\tilde{y}_{sj}` come from the same distribution as the
    measurements :math:`y_{ij}`, where :math:`s` indexes simulated individuals,
    :math:`j` time points and :math:`i` measured individuals.

    Formally the log-likelihood of the simulated observations with respect to
    the filter is defined as

    .. math::
        \log p(Y | \tilde{Y}) = \sum _{ij} \log p(y_{ij} | \tilde{Y}_j ),

    where :math:`\tilde{Y}_j = \{ \tilde{y}_{sj} \}` are the simulated
    observations at time point :math:`t_j`.

    The measurements are expected to be arranged into a 3 dimensional numpy
    array of shape ``(n_ids, n_observables, n_times)``, where
    ``n_ids`` is the number of measured individuals at a given time point,
    ``n_observables`` is the number of unique observables that were
    measurement, and ``n_times`` is the number of unique time points.
    The filter expects the simulated measurements to be ordered in
    the same way, so no record of the measurement times is needed.

    If varying numbers of individuals were measured at different time points,
    or not all observables were measured for each individual, the missing
    values can be filled with ``np.nan`` to be able to shape the observations
    into the format ``(n_ids, n_observables, n_times)``. The missing values
    will be filtered internally.

    :param observations: Measurements.
    :type observations: np.ndarray of shape
        ``(n_ids, n_observables, n_times)``
    """
    def __init__(self, observations):
        observations = np.asarray(observations)
        if observations.ndim != 3:
            raise ValueError(
                'The observations must be of shape '
                '(n_ids, n_observables, n_times).')

        # Transform into masked array
        mask = np.isnan(observations)
        if np.any(mask):
            observations = np.ma.array(observations, mask=mask)

        self._observations = observations
        _, self._n_observables, self._n_times = observations.shape

    def compute_log_likelihood(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: float
        """
        raise NotImplementedError

    def compute_sensitivities(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter, and the sensitivities of the log-likelihood
        with respect to the simulated observations.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: Tuple[float, np.ndarray] where the array has shape
            ``(n_sim, n_observables, n_times)``
        """
        raise NotImplementedError

    def n_observables(self):
        """
        Returns the number of observables in the dataset.
        """
        return self._n_observables

    def n_times(self):
        """
        Returns the number of measurement times in the dataset.
        """
        return self._n_times

    def sort_times(self, order):
        """
        Sorts the observations along the time dimension according to the
        provided indices.

        :param order: An array with indices that orders the observations along
            thetime dimension.
        :type order: np.ndarray of shape ``(n_times,)``
        """
        order = np.asarray(order)
        if len(order) != self._n_times:
            raise ValueError('Order has to be of length n_times.')
        if len(order) != len(np.unique(order)):
            raise ValueError('Order has to contain n_times unique elements.')

        self._observations = self._observations[..., order]


class ComposedPopulationFilter(PopulationFilter):
    r"""
    A filter composed of multiple filters.

    A composed filter takes a list of filters and defines
    the log-likelihood of simulated measurements as the sum over the individual
    log-likelihoods of the filters

    .. math::
        \log p(Y | \tilde{Y}) = \sum _{ij} \log p_j(y_{ij} | \tilde{Y}_j ),

    where :math:`\tilde{Y}_j = \{ \tilde{y}_{sj} \}` are the simulated
    measurements and :math:`p_j(\cdot | \tilde{Y}_j )` is the filter
    at time point :math:`t_j`.

    The input instances of :class:`chi.PopulationFilter` may model
    multiple time points at once. The measurement times are expected to be
    ordered according to the concatenated measurement times of the individual
    filters.

    Extends :class:`chi.PopulationFilter`.

    :param population_filters: List of filters.
    :type population_filters: List[chi.PopulationFilter]
    """
    def __init__(self, population_filters):
        # Check inputs
        for population_filter in population_filters:
            if not isinstance(population_filter, PopulationFilter):
                raise TypeError(
                    'All filters have to be instances '
                    'of chi.PopulationFilter.')
        # TODO: Enforce that all filters model the same number of
        # observables. In future PRs we can think about how to relax this
        # constraint (might be nice to model different observables at the same
        # time with different filters).
        n_observables = population_filters[0].n_observables()
        for population_filter in population_filters:
            if population_filter.n_observables() != n_observables:
                raise ValueError(
                    'All filters need to model the same number of '
                    'observables.')
        self._population_filters = population_filters
        self._n_observables = n_observables

        # Get properties
        self._n_times = np.sum([
            pop_filter.n_times() for pop_filter in self._population_filters])

        # Defaults
        self._time_order = None
        self._time_filter_order = None

    def compute_log_likelihood(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: float
        """
        # Sort simulated observations
        if self._time_order is not None:
            simulated_obs = simulated_obs[:, :, self._time_filter_order]

        # Compute score
        score = 0
        current_time_id = 0
        for pop_filter in self._population_filters:
            end_time_id = current_time_id + pop_filter.n_times()
            score += pop_filter.compute_log_likelihood(
                simulated_obs[:, :, current_time_id:end_time_id])
            current_time_id = end_time_id

        return score

    def compute_sensitivities(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter, and the sensitivities of the log-likelihood
        with respect to the simulated observations.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: Tuple[float, np.ndarray] where the array has shape
            ``(n_sim, n_observables, n_times)``
        """
        # Sort simulated observations
        if self._time_order is not None:
            simulated_obs = simulated_obs[:, :, self._time_filter_order]

        # Compute score
        score = 0
        sensitivities = np.zeros(shape=simulated_obs.shape)
        current_time_id = 0
        for pop_filter in self._population_filters:
            end_time_id = current_time_id + pop_filter.n_times()
            s, sens = pop_filter.compute_sensitivities(
                simulated_obs[:, :, current_time_id:end_time_id])
            score += s
            sensitivities[:, :, current_time_id:end_time_id] = sens
            current_time_id = end_time_id

        # Sort sensitivities into input order
        if self._time_order is not None:
            sensitivities = sensitivities[:, :, self._time_order]

        return score, sensitivities

    def n_observables(self):
        """
        Returns the number of observables in the dataset.
        """
        return self._n_observables

    def n_times(self):
        """
        Returns the number of measurement times in the dataset.
        """
        return self._n_times

    def sort_times(self, order):
        """
        Sorts the observations along the time dimension according to the
        provided indices.

        :param order: An array with indices that orders the observations along
            thetime dimension.
        :type order: np.ndarray of shape ``(n_times,)``
        """
        order = np.asarray(order)
        if len(order) != self._n_times:
            raise ValueError('Order has to be of length n_times.')
        if len(order) != len(np.unique(order)):
            raise ValueError('Order has to contain n_times unique elements.')

        # If order doesn't change, ignore input
        # (This is purely for efficiency of log-likelihood computation)
        if np.all(order == np.arange(self._n_times)):
            return None

        # Cannot sort observations directly
        # (compartmentalised into individual filters)
        # So let us remember the order and order simulated measurements
        # during inference
        self._time_order = np.copy(order)
        self._time_filter_order = np.argsort(self._time_order)


class GaussianFilter(PopulationFilter):
    r"""
    Implements a Gaussian filter.

    A Gaussian filter approximates the distribution of
    measurements at time point :math:`t_j` by a Gaussian distribution
    whose mean and variance are estimated from simulated
    measurements

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ij} \log \mathcal{N} (y_{ij} | \mu _j, \sigma ^2_j),

    where the mean :math:`\mu _j` and the variance :math:`\sigma ^2_j` are
    given by empirical estimates from the simulated measurements

    .. math::
        \mu _j = \frac{1}{n_s} \sum _{s=1}^{n_s} \tilde{y}_{sj}
        \quad \text{and} \quad
        \sigma ^2 _j = \frac{1}{n_s-1} \sum _{s=1}^{n_s} \left(
            \tilde{y}_{sj} - \mu _j \right) ^2.

    Here, we use :math:`i` to index measured individuals from the dataset,
    :math:`j` to index measurement time points and :math:`s` to index simulated
    measurements. :math:`n_s` denotes the number of simulated measurements per
    time point.

    For multiple measured observables the above expression can be
    straightforwardly extended to

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ijr} \log \mathcal{N} (y_{ijr} | \mu _{jr}, \sigma ^2_{jr}),

    where :math:`r` indexes observables and :math:`\mu _{jr}` and
    :math:`\sigma^2 _{jr}` are the empirical mean and variance over the
    simulated measurements of the observable :math:`r` at time point
    :math:`t_j`.

    Extends :class:`PopulationFilter`

    :param observations: Measurements.
    :type observations: np.ndarray of shape
        ``(n_ids, n_observables, n_times)``
    """
    def __init__(self, observations):
        super().__init__(observations)

    def _compute_log_likelihood(self, mu, var):
        """
        Returns the log-likelihood.

        mu of shape (1, n_observables, n_times)
        var of shape (1, n_observables, n_times)
        """
        score = -np.sum(
            np.log(2*np.pi) + np.log(var)
            + (self._observations - mu)**2 / var) / 2

        return score

    def compute_log_likelihood(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: float
        """
        mu = np.mean(simulated_obs, axis=0, keepdims=True)
        var = np.var(simulated_obs, ddof=1, axis=0, keepdims=True)

        score = self._compute_log_likelihood(mu, var)
        if np.ma.is_masked(score):
            return -np.inf

        return score

    def compute_sensitivities(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter, and the sensitivities of the log-likelihood
        with respect to the simulated observations.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: Tuple[float, np.ndarray] where the array has shape
            ``(n_sim, n_observables, n_times)``
        """
        mu = np.mean(simulated_obs, axis=0, keepdims=True)
        var = np.var(simulated_obs, ddof=1, axis=0, keepdims=True)

        score = self._compute_log_likelihood(mu, var)
        if np.ma.is_masked(score):
            return -np.inf, np.empty(simulated_obs.shape)

        # Compute sensitivities
        # dscore/dsim = dscore/dmu dmu/dsim + dscore/dvar dvar/dsim
        n_sim = len(simulated_obs)
        dscore_dsim = \
            np.sum((self._observations - mu) / var, axis=0) / n_sim \
            + np.sum(
                - 1 / var + (self._observations - mu)**2 / var**2,
                axis=0) * (simulated_obs - mu) / (n_sim - 1)

        return score, dscore_dsim


class GaussianKDEFilter(PopulationFilter):
    r"""
    Implements a Gaussian kernel density estimation filter.

    A Gaussian KDE filter approximates the distribution of
    measurements at time point :math:`t_j` by a Gaussian KDE
    approximation of the simulated measurements. The Gaussian KDE approximation
    is defined by the average over Gaussian probability densities whose means
    are equal to the simulated measurements and the standard deviation (or
    bandwidth) is a hyperparameter. By default the bandwidth is chosen by
    an adapted rule of thumb.

    The log-likelihood of the simulated measurements with respect to the
    measurements and the filter is defined as

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ij} \log \left( \frac{1}{n_s} \sum _{s=1}^{n_s}
            \mathcal{N} (y_{ij} | \tilde{y}_{sj}, \tilde{\sigma} ^2_j) \right).

    Here, we use :math:`i` to index measured individuals from the dataset,
    :math:`j` to index measurement time points and :math:`s` to index simulated
    measurements. :math:`n_s` denotes the number of simulated measurements per
    time point.

    An adapted rule of thumb is used to estimate an
    appropriate bandwidth for each time point :math:`t_j`

    .. math::
        \tilde{\sigma} _j =
            \left( \frac{4}{3n_s}\right) ^ {1/5}
            \sqrt{\frac{1}{n_s - 1}
            \sum _s (\tilde{y}_{sj} - \tilde{\mu} _j)^2},

    where :math:`\tilde{\mu} _j = \sum _s \tilde{y}_{sj} / n_s` is the
    empirical mean over the simulated measurements at time
    :math:`t_j`.

    For multiple measured observables the above expression can be
    straightforwardly extended to

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ijr} \log \left( \frac{1}{n_s} \sum _{s=1}^{n_s}
            \mathcal{N} (y_{ijr} | \tilde{y}_{sjr}, \tilde{\sigma} ^2_{jr})
            \right),

    where :math:`r` indexes observables and
    :math:`\tilde{\sigma} _{jr}` is the bandwidth for observable :math:`r` at
    time point :math:`t_j`.

    Extends :class:`PopulationFilter`

    :param observations: Measurements.
    :type observations: np.ndarray of shape
        ``(n_ids, n_observables, n_times)``
    """
    def __init__(self, observations):
        super().__init__(observations)

        # Add dummy dimension to observations for later convenience
        self._observations = self._observations[np.newaxis, ...]

    def compute_log_likelihood(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: float
        """
        n_sim = len(simulated_obs)
        simulated_obs = np.asarray(simulated_obs)[:, np.newaxis, :, :]
        bw_squared = (4 / 3 / n_sim) ** 0.4 * np.var(
            simulated_obs, ddof=1, axis=0, keepdims=True)

        score = np.sum(logsumexp(
            - (simulated_obs - self._observations)**2
            / bw_squared / 2, axis=0
            ) - np.log(n_sim) - np.log(2 * np.pi) / 2 - np.log(bw_squared) / 2)
        if np.ma.is_masked(score):
            return -np.inf

        return score

    def compute_sensitivities(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter, and the sensitivities of the log-likelihood
        with respect to the simulated observations.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: Tuple[float, np.ndarray] where the array has shape
            ``(n_sim, n_observables, n_times)``
        """
        n_sim = len(simulated_obs)
        simulated_obs = np.asarray(simulated_obs)[:, np.newaxis, :, :]
        mean = np.mean(simulated_obs, axis=0, keepdims=True)
        var = np.var(simulated_obs, ddof=1, axis=0, keepdims=True)
        bw_squared = (4 / 3 / n_sim) ** 0.4 * var

        # Compute log-likelihood
        scores = \
            - (simulated_obs - self._observations)**2 / bw_squared / 2
        score = np.sum(
            logsumexp(scores, axis=0) - np.log(n_sim) - np.log(2 * np.pi) / 2
            - np.log(bw_squared) / 2)
        if np.ma.is_masked(score):
            n_sim, _, n_obs, n_times = simulated_obs.shape
            return -np.inf, np.empty((n_sim, n_obs, n_times))

        # Compute sensitivities
        # score = log mean exp scores - log bw + constant
        # dscore/dsim =
        #   exp(scores) / sum(exp(scores)) * dscores / dsim
        #   - 1 / bw^2 / 2 * dbw^2 / dsim
        dbw_squared_dsim_by_bw_squared = \
            2 * (simulated_obs - mean) / (n_sim - 1) / var
        dscore_dsim = np.sum(
            softmax(scores, axis=0)
            * (self._observations - simulated_obs) / bw_squared
            - np.sum(softmax(scores, axis=0) * scores, axis=0, keepdims=True)
            * dbw_squared_dsim_by_bw_squared
            - dbw_squared_dsim_by_bw_squared / 2, axis=1)

        return score, dscore_dsim


class GaussianMixtureFilter(PopulationFilter):
    r"""
    Implements a Gaussian mixture filter.

    A Gaussian mixture filter approximates the distribution of
    measurements at time point :math:`t_j` by a Gaussian mixture distribution
    whose kernel means and variances are estimated from simulated
    measurements

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ij} \log \sum_m \frac{1}{M}
            \mathcal{N} (y_{ij} | \mu _{jm}, \sigma ^2_{jm}),

    where the mean :math:`\mu _{jm}` and the variance :math:`\sigma ^2_{jm}`
    of the mth Gaussian distribution are given by the empirical estimates from
    the mth subset of the simulated measurements

    .. math::
        \mu _{jm} = \frac{1}{n} \sum _{s=1}^{n} \tilde{y}_{sjm}
        \quad \text{and} \quad
        \sigma ^2 _{jm} = \frac{1}{n-1} \sum _{s=1}^{n} \left(
            \tilde{y}_{sjm} - \mu _{jm} \right) ^2.

    Here, we use :math:`i` to index measured individuals from the dataset,
    :math:`j` to index measurement time points and :math:`s` to index simulated
    measurements. :math:`n` denotes the number of simulated measurements per
    time point.

    For multiple measured observables the above expression can be
    straightforwardly extended to

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ijr} \log \sum_m \frac{1}{M}
            \mathcal{N} (y_{ijr} | \mu _{jrm}, \sigma ^2_{jrm}),

    where :math:`r` indexes observables and :math:`\mu _{jrm}` and
    :math:`\sigma^2 _{jrm}` are the empirical mean and variance over the
    mth subset of simulated measurements of the observable :math:`r` at time
    point :math:`t_j`.

    Extends :class:`PopulationFilter`

    :param observations: Measurements.
    :type observations: np.ndarray of shape
        ``(n_ids, n_observables, n_times)``
    """
    def __init__(self, observations, n_kernels=2):
        super().__init__(observations)
        n_kernels = int(n_kernels)
        if n_kernels < 2:
            raise ValueError(
                'Invalid number of kernels. A Gaussian mixture filter expects '
                'at least 2 kernels.')
        self._n_kernels = n_kernels

        # Add dummy dimension to observations for later convenience
        self._observations = self._observations[
            np.newaxis, np.newaxis, :, :, :]

    def _compute_log_likelihood(self, mu, var):
        """
        Returns the log-likelihood.

        mu of shape (n_kernels, 1, 1, n_observables, n_times)
        var of shape (n_kernels, 1, 1, n_observables, n_times)
        """
        score = np.sum(logsumexp(
            - (mu - self._observations)**2 / var / 2 - np.log(var) / 2, axis=0)
            - np.log(self._n_kernels) - np.log(2 * np.pi) / 2)
        if np.ma.is_masked(score):
            return -np.inf

        return score

    def compute_log_likelihood(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: float
        """
        if len(simulated_obs) % self._n_kernels > 0:
            raise ValueError(
                'Invalid simulated_obs. The number of simulated observations '
                'needs to be a multiple of the number of kernels.')

        # Compute means and variances
        n_sim, n_obs, n_times = simulated_obs.shape
        n_per_kernel = n_sim // self._n_kernels
        simulated_obs = simulated_obs.reshape(
            self._n_kernels, n_per_kernel, 1, n_obs, n_times)
        mu = np.mean(simulated_obs, axis=1, keepdims=True)
        var = np.var(simulated_obs, ddof=1, axis=1, keepdims=True)

        score = self._compute_log_likelihood(mu, var)

        return score

    def compute_sensitivities(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter, and the sensitivities of the log-likelihood
        with respect to the simulated observations.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: Tuple[float, np.ndarray] where the array has shape
            ``(n_sim, n_observables, n_times)``
        """
        if len(simulated_obs) % self._n_kernels > 0:
            raise ValueError(
                'Invalid simulated_obs. The number of simulated observations '
                'needs to be a multiple of the number of kernels.')

        # Compute means and variances
        n_sim, n_obs, n_times = simulated_obs.shape
        n_per_kernel = n_sim // self._n_kernels
        simulated_obs = simulated_obs.reshape(
            self._n_kernels, n_per_kernel, 1, n_obs, n_times)
        mu = np.mean(simulated_obs, axis=1, keepdims=True)
        var = np.var(simulated_obs, ddof=1, axis=1, keepdims=True)

        # Compute log-likelihood
        scores = - (mu - self._observations)**2 / var / 2 - np.log(var) / 2
        score = np.sum(
            logsumexp(scores, axis=0) - np.log(self._n_kernels)
            - np.log(2 * np.pi) / 2)
        if np.ma.is_masked(score):
            return -np.inf, np.empty((n_sim, n_obs, n_times))

        # Compute sensitivities
        # score = log mean exp scores + constant
        # dscore/dsim =
        #   exp(scores) / sum(exp(scores)) * dscores / dsim
        dscores_dsim = \
            (self._observations - mu) / var / n_per_kernel \
            + (- 1 / var + (self._observations - mu)**2 / var**2) \
            * (simulated_obs - mu) / (n_per_kernel - 1)
        dscore_dsim = np.sum(softmax(scores, axis=0) * dscores_dsim, axis=2)
        dscore_dsim = dscore_dsim.reshape(n_sim, n_obs, n_times)

        return score, dscore_dsim


class LogNormalFilter(PopulationFilter):
    r"""
    Implements a lognormal filter.

    A lognormal filter approximates the distribution of
    measurements at time point :math:`t_j` by a lognormal distribution
    whose location and scale are estimated from simulated
    measurements

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ij} \log \mathrm{LN} (y_{ij} | \mu _j, \sigma _j),

    where the location :math:`\mu _j` and the location :math:`\sigma _j` are
    given by empirical estimates of the log-mean and the log-standard deviation
    of the simulated measurements

    .. math::
        \mu _j = \frac{1}{n_s} \sum _{s=1}^{n_s} \log \tilde{y}_{sj}
        \quad \text{and} \quad
        \sigma _j = \sqrt{\frac{1}{n_s-1} \sum _{s=1}^{n_s} \left(
            \log \tilde{y}_{sj} - \mu _j \right) ^2}.

    Here, we use :math:`i` to index measured individuals from the dataset,
    :math:`j` to index measurement time points and :math:`s` to index simulated
    measurements. :math:`n_s` denotes the number of simulated measurements per
    time point.

    For multiple measured observables the above expression can be
    straightforwardly extended to

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ijr} \log \mathrm{LN} (y_{ijr} | \mu _{jr}, \sigma _{jr}),

    where :math:`r` indexes observables and :math:`\mu _{jr}` and
    :math:`\sigma _{jr}` are the empirical mean and variance over the
    simulated log-measurements of the observable :math:`r` at time point
    :math:`t_j`.

    Extends :class:`PopulationFilter`

    :param observations: Measurements.
    :type observations: np.ndarray of shape
        ``(n_ids, n_observables, n_times)``
    """
    def __init__(self, observations):
        super().__init__(observations)

        # Log-transform the observations for later convenience
        self._observations = np.log(self._observations)

    def _compute_log_likelihood(self, mu, var):
        """
        Returns the log-likelihood.

        mu of shape (1, n_observables, n_times)
        var of shape (1, n_observables, n_times)
        """
        score = -np.sum(
            np.log(2*np.pi) + np.log(var) + self._observations
            + (self._observations - mu)**2 / var) / 2

        return score

    def compute_log_likelihood(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: float
        """
        simulated_obs = np.log(simulated_obs)
        mu = np.mean(simulated_obs, axis=0, keepdims=True)
        var = np.var(simulated_obs, ddof=1, axis=0, keepdims=True)

        score = self._compute_log_likelihood(mu, var)
        if np.ma.is_masked(score):
            return -np.inf

        return score

    def compute_sensitivities(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter, and the sensitivities of the log-likelihood
        with respect to the simulated observations.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: Tuple[float, np.ndarray] where the array has shape
            ``(n_sim, n_observables, n_times)``
        """
        simulated_obs = np.asarray(simulated_obs)
        log_simulated_obs = np.log(simulated_obs)
        mu = np.mean(log_simulated_obs, axis=0, keepdims=True)
        var = np.var(log_simulated_obs, ddof=1, axis=0, keepdims=True)

        score = self._compute_log_likelihood(mu, var)
        if np.ma.is_masked(score):
            return -np.inf, np.empty(simulated_obs.shape)

        # Compute sensitivities
        # dscore/dsim = dscore/dmu dmu/dsim + dscore/dvar dvar/dsim
        n_sim = len(simulated_obs)
        dscore_dsim = (
            np.sum((self._observations - mu) / var, axis=0) / n_sim
            + np.sum(
                (self._observations - mu)**2 / var**2 - 1 / var, axis=0
            ) / (n_sim - 1) * ((log_simulated_obs - mu) - np.mean(
                log_simulated_obs - mu, axis=0))
        ) / simulated_obs

        return score, dscore_dsim


class LogNormalKDEFilter(PopulationFilter):
    r"""
    Implements a lognormal kernel density estimation filter.

    A lognormal KDE filter approximates the distribution of
    measurements at time point :math:`t_j` by a lognormal KDE
    approximation of the simulated measurements. The lognormal KDE
    approximation is defined by the average over lognormal probability
    densities whose locations are equal to the simulated measurements and the
    scale (or bandwidth) is a hyperparameter. By default the bandwidth is
    chosen by an adapted rule of thumb.

    The log-likelihood of the simulated measurements with respect to the
    measurements and the filter is defined as

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ij} \log \left( \frac{1}{n_s} \sum _{s=1}^{n_s}
            \mathrm{LN} (y_{ij} | \tilde{y}_{sj}, \sigma _j) \right).

    Here, we use :math:`i` to index measured individuals from the dataset,
    :math:`j` to index measurement time points and :math:`s` to index simulated
    measurements. :math:`n_s` denotes the number of simulated measurements per
    time point.

    An adapted rule of thumb is used to estimate an
    appropriate bandwidth for each time point :math:`t_j`

    .. math::
        \sigma _j =
            \left( \frac{4}{3n_s}\right) ^ {1/5}
            \sqrt{\frac{1}{n_j - 1}\sum _i (\log y_{ij} - \mu _j)^2},

    where :math:`\mu _j = \sum _i \log y_{ij} / n_j` is the empirical mean
    over the log-measurements and :math:`n_j` is the number of measurements at
    time :math:`t_j`. Note that this deviates from the standard definition of
    the rule of thumb, where the empirical variance would be estimated from the
    simulated measurements.

    For multiple measured observables the above expression can be
    straightforwardly extended to

    .. math::
        \log p(\mathcal{D} | \tilde{Y}) =
            \sum _{ijr} \log \left( \frac{1}{n_s} \sum _{s=1}^{n_s}
            \mathrm{LN} (y_{ijr} | \tilde{y}_{sjr}, \sigma _{jr}) \right),

    where :math:`r` indexes observables and
    :math:`\sigma _{jr}` is the bandwidth for observable :math:`r` at time
    point :math:`t_j`.

    Extends :class:`PopulationFilter`

    :param observations: Measurements.
    :type observations: np.ndarray of shape
        ``(n_ids, n_observables, n_times)``
    """
    def __init__(self, observations, bandwidth=None):
        super().__init__(observations)

        # Log-transform and reshape for later convenience
        self._observations = np.log(self._observations)[np.newaxis, ...]

    def compute_log_likelihood(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: float
        """
        n_sim = len(simulated_obs)
        simulated_obs = np.log(np.asarray(simulated_obs))[:, np.newaxis, :, :]
        bw_squared = (4 / 3 / n_sim) ** 0.4 * np.var(
            simulated_obs, ddof=1, axis=0, keepdims=True)

        score = np.sum(logsumexp(
            - (simulated_obs - self._observations)**2
            / bw_squared / 2, axis=0
            ) - np.log(n_sim) - np.log(2 * np.pi) / 2 - np.log(bw_squared) / 2)
        if np.ma.is_masked(score):
            return -np.inf

        return score

    def compute_sensitivities(self, simulated_obs):
        """
        Returns the log-likelihood of the simulated observations with respect
        to the data and filter, and the sensitivities of the log-likelihood
        with respect to the simulated observations.

        :param simulated_obs: Simulated measurements.
        :type simulated_obs: np.ndarray of shape
            (n_sim, n_observables, n_times)
        :rtype: Tuple[float, np.ndarray] where the array has shape
            ``(n_sim, n_observables, n_times)``
        """
        n_sim = len(simulated_obs)
        simulated_obs = np.asarray(simulated_obs)[:, np.newaxis, :, :]
        log_simulated_obs = np.log(simulated_obs)
        mean = np.mean(log_simulated_obs, axis=0, keepdims=True)
        var = np.var(log_simulated_obs, ddof=1, axis=0, keepdims=True)
        bw_squared = (4 / 3 / n_sim) ** 0.4 * var

        # Compute log-likelihood
        scores = \
            - (log_simulated_obs - self._observations)**2 \
            / bw_squared / 2
        score = np.sum(
            logsumexp(scores, axis=0) - np.log(n_sim) - np.log(2 * np.pi) / 2
            - np.log(bw_squared) / 2)
        if np.ma.is_masked(score) or np.isnan(score):
            n_sim, _, n_obs, n_times = simulated_obs.shape
            return -np.inf, np.empty((n_sim, n_obs, n_times))

        # Compute sensitivities
        # score = log mean exp scores - log bw + constant
        # dscore/dsim =
        #   exp(scores) / sum(exp(scores)) * dscores / dsim
        #   - 1 / bw^2 / 2 * dbw^2 / dsim
        dbw_squared_dsim_by_bw_squared = \
            2 * (log_simulated_obs - mean) / (n_sim - 1) / var / simulated_obs
        dscore_dsim = np.sum(
            softmax(scores, axis=0)
            * (self._observations - log_simulated_obs) / bw_squared
            / simulated_obs
            - np.sum(softmax(scores, axis=0) * scores, axis=0, keepdims=True)
            * dbw_squared_dsim_by_bw_squared
            - dbw_squared_dsim_by_bw_squared / 2, axis=1)

        return score, dscore_dsim


def logsumexp(a, axis=None, keepdims=False):  # pragma: no cover
    """
    Returns log of the sum of the exponentiated values of a.

    Code is copied from scipy.special.logsumexp and adapted to work for masked
    arrays.
    """
    a_max = np.max(a, axis=axis, keepdims=True)

    if a_max.ndim > 0:
        a_max[~np.isfinite(a_max)] = 0
    elif not np.isfinite(a_max):
        a_max = 0

    tmp = np.exp(a - a_max)

    # suppress warnings about log of zero
    with np.errstate(divide='ignore'):
        s = np.sum(tmp, axis=axis, keepdims=keepdims)
        out = np.log(s)

    # Add back shift
    if not keepdims:
        a_max = np.squeeze(a_max, axis=axis)
    out += a_max

    return out


def softmax(a, axis=None):  # pragma: no cover
    """
    Returns the softmax of a.

    Code is copied from scipy.special.softmax and adapted to work for masked
    arrays.
    """
    return np.exp(a - logsumexp(a, axis=axis, keepdims=True))



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import arviz as az
import numpy as np
import pandas as pd
import pints
from tqdm.notebook import tqdm
import xarray as xr

import chi


# TODO: Revisit pointwise log-likelihood evaluation for population models
def _compute_hierarchical_pointwise_log_likelihood(
        log_likelihood, posterior, chain_coords, draw_coords,
        per_individual, show_chain_progress_bar):  # pragma: no cover
    """
    Computes the pointwise log-likelihoods for each sample in the
    posterior for a hierarchical model.

    If per individual is ``True`` the pointwise log-likelihoods are
    aggregated per individual. Otherwise the pointwise log-likelihoods
    are computed for each observation.
    """
    # Create container for pointwise log-likelihoods
    # (either per individual or per observation)
    n_chains, n_draws, _ = posterior.shape
    if per_individual is True:
        n_ids = log_likelihood.n_log_likelihoods()
        pointwise_ll = np.empty(shape=(n_chains, n_draws, n_ids))
    else:
        n_obs = np.sum(log_likelihood.n_observations())
        pointwise_ll = np.empty(shape=(n_chains, n_draws, n_obs))

    for chain_id, chain in enumerate(tqdm(
            posterior, disable=not show_chain_progress_bar)):
        for draw_id, draw in enumerate(chain):
            # Compute pointwise log-likelihood for the
            # given iteration and chain
            pointwise_ll[chain_id, draw_id, :] = \
                log_likelihood.compute_pointwise_ll(
                    draw, per_individual)

    # Annotate as xarray
    individual = log_likelihood.get_id(individual_ids=True)
    if per_individual:
        pointwise_ll = xr.DataArray(
            data=pointwise_ll,
            coords={
                'chain': chain_coords,
                'draw': draw_coords,
                'individual': individual},
            dims=('chain', 'draw', 'individual'))
    else:
        obs_coords = []
        for index, n_obs in enumerate(log_likelihood.n_observations()):
            _id = individual[index]
            obs_coords += [
                '%s Observation %d' % (_id, obs) for obs in range(1, n_obs+1)]
        pointwise_ll = xr.DataArray(
            data=pointwise_ll,
            coords={
                'chain': chain_coords,
                'draw': draw_coords,
                'observation': obs_coords},
            dims=('chain', 'draw', 'observation'))

    return pointwise_ll


def _compute_individual_pointwise_log_likelihood(
        log_likelihood, posterior, chain_coords, draw_coords,
        show_chain_progress_bar):  # pragma: no cover
    """
    Computes the pointwise log-likelihoods for each sample in the
    posterior for a non-hierarchical model.
    """
    # Create container for pointwise log-likelihoods
    # (either per individual or per observation)
    n_obs = np.sum(log_likelihood.n_observations())
    n_chains, n_draws, _ = posterior.shape
    pointwise_ll = np.empty(shape=(n_chains, n_draws, n_obs))

    for chain_id, chain in enumerate(tqdm(
            posterior, disable=not show_chain_progress_bar)):
        for draw_id, draw in enumerate(chain):
            # Compute pointwise log-likelihood for the
            # given iteration and chain
            pointwise_ll[chain_id, draw_id, :] = \
                log_likelihood.compute_pointwise_ll(draw)

    # Annotate as xarray
    obs_coords = []
    for out_id, n_obs in enumerate(log_likelihood.n_observations()):
        output_id = out_id + 1
        obs_coords += [
            'Output %d Observation %d' % (output_id, obs)
            for obs in range(1, n_obs+1)]
    pointwise_ll = xr.DataArray(
        data=pointwise_ll,
        coords={
            'chain': chain_coords,
            'draw': draw_coords,
            'observation': obs_coords},
        dims=('chain', 'draw', 'observation'))

    return pointwise_ll


def _check_parameters(
        log_likelihood, posterior_samples, param_map,
        is_hierarchical):  # pragma: no cover
    """
    Checks that all log-likelihood parameters can be found in the
    posterior.

    For individual log-likelihoods the parameter names must coincide
    with variables of the posterior xr.Dataset, or the names can be
    mapped witg the param_map.

    For hierarchical log-likelihoods the top-level parameters can
    be equally identified. However the bottom level parameters are
    more complicated:

    The posterior xr.Dataset will have a variable <parameter name>
    which has the individuals' IDs as a dimension. So to check the
    existence of the bottom parameters we first check for the
    parameter name and then the ID.
    """
    # Create map from posterior parameter names to model parameter names
    # (This map applies equally to top and bottom parameters - IDs excluded)
    model_names = log_likelihood.get_parameter_names()
    for param_id, name in enumerate(model_names):
        try:
            model_names[param_id] = param_map[name]
        except KeyError:
            # The name is not mapped
            pass

    # Check that model names are in the dataset
    for parameter in model_names:
        if parameter not in posterior_samples.data_vars:
            raise ValueError(
                'The parameter <' + str(parameter) + '> cannot be found '
                'in the posterior.')

    # Also get the mapped top-parameters
    # (For non-hierarchical models top_parameters are the same as
    # model_names)
    top_parameters = model_names
    if is_hierarchical is True:
        top_parameters = log_likelihood.get_parameter_names(
            exclude_bottom_level=True)

    # Filter heterogenous parameters from top-level parameters
    # (In posterior Dataset they are treated as bottom level parameters)
    top_parameters = _filter_top_parameters(top_parameters)

    # For a hierarchical log-likelihood make sure that bottom parameters
    # also exist for all IDs
    if is_hierarchical is True:
        ids = log_likelihood.get_id(individual_ids=True)
        for parameter in model_names:
            # Skip if parameter is top level
            if parameter in top_parameters:
                continue

            # Check that IDs exist
            param_ids = posterior_samples[parameter].individual
            for _id in ids:
                if _id not in param_ids:
                    raise ValueError(
                        'The ID <' + str(_id) + '> does not exist in the '
                        'posterior for parameter <' + str(parameter) + '>.')

    return model_names, top_parameters


def _filter_top_parameters(top_parameters):  # pragma: no cover
    """
    Filter heterogenously modelled top parameters from the list.

    For the formatting of the hierarchical posterior samples we need to
    distinguish between top and bottom level parameters.

    Top parameters are parameters of the population model, bottom parameters
    are parameters of the individuals. However, top and bottom level
    parameters are not mutually exclusive, since for a HeterogeneousModel
    the individual parameters are also top level parameters.

    In the posterior Dataset parameters that are associated to individuals
    have an individual coordinate, while parameters that are associated to
    the population do not have an individual coordinate.

    Almost all top-level parameters will fall into the class of parameters
    without individual coordinates, except HeterogenousModel parameters.
    Hence, we filter them from the top-level parameter list, such that they
    are treated as individual parameters for the formatting.
    """
    top_params = []
    top_parameters = top_parameters.copy()
    while top_parameters:
        # Keep parameter if it only occurs once
        # (else it must be a heterogeneously modelled parameter)
        parameter = top_parameters.pop(0)
        if parameter not in top_parameters:
            top_params.append(parameter)

        # Remove all remaining occurences
        top_parameters = [p for p in top_parameters if p != parameter]

    return top_params


def _format_posterior(
        parameter_names, top_parameters, posterior_samples,
        ids):  # pragma: no cover
    """
    Formats the dataset into a numpy array of shape
    (n_iterations, n_parameters), such that looping through the iterations
    can be done more efficiently.
    """
    n_chains = len(posterior_samples.chain)
    n_parameters = len(parameter_names)
    n_draws = len(posterior_samples.dropna(dim='draw').draw)

    # Collect samples
    id_index = 0
    n_ids = len(ids) if ids is not None else 1
    posterior = np.empty(shape=(n_chains, n_draws, n_parameters))
    for param_id, parameter in enumerate(parameter_names):
        if parameter in top_parameters:
            try:
                # Assumes values have shape (n_chains, n_draws)
                posterior[:, :, param_id] = \
                    posterior_samples[parameter].values
            except ValueError:
                # If broadcasting fails, values must be transposed
                posterior[:, :, param_id] = \
                    posterior_samples[parameter].values.T

            # Continue to next parameter
            continue  # pragma: no cover

        # Get parameters of individual
        _id = ids[id_index % n_ids]
        try:
            # Assumes values have shape (n_chains, n_draws)
            posterior[:, :, param_id] = \
                posterior_samples[parameter].sel(individual=_id).values
        except ValueError:
            # If broadcasting fails, values must be transposed
            posterior[:, :, param_id] = \
                posterior_samples[parameter].sel(individual=_id).values.T

        # Shift id index to next individual
        # (This crucially relies on the fact that individual parameters of
        # one type are grouped together, i.e. we iterate first through all
        # individuals before moving on to the next parameter.)
        id_index += 1

    return posterior


def compute_pointwise_loglikelihood(
        log_likelihood, posterior_samples, individual=None, param_map=None,
        per_individual=True, return_inference_data=False,
        show_chain_progress_bar=False):  # pragma: no cover
    """
    Computes the pointwise log-likelihood for each observation and each
    parameter sample from the posterior distribution.

    For a :class:`HierarchicalLogLikelihood` pointwise log-likelihoods are
    by default computed and aggregated per individual. If the pointwise
    log-likelihoods are supposed to be computed per observation,
    ``per_individual`` can be set to ``False``. For more info see
    :meth:`HierarchicalLogLikelihood.compute_pointwise_ll`.

    :param log_likelihood: The log-likelihood of the model parameters.
    :type log_likelihood: LogLikelihood, HierarchicalLogLikelihood
    :param posterior_samples: Samples from the posterior distribution of the
        model parameters.
    :type posterior_samples: xarray.Dataset
    :param individual: The individual for which the log-likelihoods are
        evaluated. If ``None`` the first individual is chosen.
    :type individual: str, optional
    :param param_map: A dictionary which can be used to map log-likelihood
        parameter names to the parameter names in the :class:`xarray.Dataset`.
        If ``None``, it is assumed that the names are identical. For
        hierarchical models top and bottom names can be mapped, IDs excluded.
    :type param_map: dict, optional
    :param per_individual: A boolean flag that determines whether the
            scores are computed per individual or per observation.
    :type per_individual: bool, optional
    :param return_inference_data: A boolean flag which determines whether the
        log-likelihoods and the posterior are returned as
        :class:`arviz.InferenceData`.
    :type return_inference_data: bool, optional
    :param show_chain_progress_bar: A boolean flag which determines whether the
        progress for each chain is visualised as a progress bar.
    :type show_chain_progress_bar: bool, optional
    """
    # Check inputs
    if not isinstance(
            log_likelihood,
            (chi.LogLikelihood, chi.HierarchicalLogLikelihood)):
        raise TypeError(
            'The log-likelihood must be either an instance of a '
            'chi.LogLikelihood or a '
            'chi.HierarchicalLogLikelihood.')
    if not isinstance(posterior_samples, xr.Dataset):
        raise TypeError(
            'The posterior samples must be an instance of a '
            'xarray.Dataset.')

    dims = sorted(list(posterior_samples.dims))
    expected_dims = ['chain', 'draw', 'individual']
    if (len(dims) == 2):
        expected_dims = ['chain', 'draw']
    for dim in expected_dims:
        if dim not in dims:
            raise ValueError(
                'The posterior samples must have the dimensions '
                '(chain, draw, individual). The current dimensions are <'
                + str(dims) + '>.')

    # Set default parameter map (no mapping)
    if param_map is None:
        param_map = {}

    try:
        param_map = dict(param_map)
    except (TypeError, ValueError):
        raise TypeError(
            'The parameter map has to be convertable to a python '
            'dictionary.')

    # Check whether model is hierarchical
    is_hierarchical = False
    if isinstance(log_likelihood, chi.HierarchicalLogLikelihood):
        is_hierarchical = True

    # Check individual for population model
    if is_hierarchical is True:
        if individual is not None:
            raise ValueError(
                "Individual IDs cannot be selected for a "
                "chi.HierarchicalLogLikelihood. To evaluate compute "
                "the pointwise for an individual separately, please use "
                "construct a log-likelihood for the individual alone.")

    # Check individual exists
    else:
        ids = posterior_samples.individual
        if individual is None:
            # Get default
            individual = str(ids.data[0])
        if individual not in ids:
            raise ValueError(
                'The individual <' + str(individual) + '> could not be '
                'found in the "individual" dimension.')

        # Get relevant samples and forget about individual
        posterior_samples = posterior_samples.sel(individual=individual)

    # Check parameters
    # (For individual log-likelihoods, top_parameters are identical to
    # parameter_names)
    parameter_names, top_parameters = _check_parameters(
        log_likelihood, posterior_samples, param_map, is_hierarchical)

    # Sort parameters into numpy array for simplified iteration
    ids = log_likelihood.get_id(individual_ids=True) if isinstance(
        log_likelihood, chi.HierarchicalLogLikelihood) else None
    posterior = _format_posterior(
        parameter_names, top_parameters, posterior_samples, ids)

    # Compute pointwise log-likelihoods
    chain_coords = posterior_samples.chain
    draw_coords = posterior_samples.draw
    if is_hierarchical is True:
        pointwise_ll = _compute_hierarchical_pointwise_log_likelihood(
            log_likelihood, posterior, chain_coords, draw_coords,
            per_individual, show_chain_progress_bar)
    else:
        pointwise_ll = _compute_individual_pointwise_log_likelihood(
            log_likelihood, posterior, chain_coords, draw_coords,
            show_chain_progress_bar)

    if return_inference_data is False:
        return pointwise_ll

    # Compose posterior samples and pointwise log-likelihood to
    # arviz.InferenceData
    samples_stats = xr.Dataset({'log_likelihood': pointwise_ll})
    inference_data = az.InferenceData(
        posterior=posterior_samples,
        sample_stats=samples_stats)

    return inference_data


class InferenceController(object):
    """
    A base class for inference controllers.

    :param log_posterior: The log-posterior from which is sampled.
    :type log_posterior: chi.LogPosterior, chi.HierarchicalLogPosterior
    :param seed: Seed for the random initialisation. Initial points are sampled
        from the log-prior.
    :type seed: int, optional
    """

    def __init__(self, log_posterior, seed=None):
        super(InferenceController, self).__init__()

        # Convert log-posterior to a list of log-posteriors
        if not isinstance(
                log_posterior,
                (chi.LogPosterior, chi.HierarchicalLogPosterior)):
            raise ValueError(
                'The log-posterior has to be an instance of a '
                'chi.LogPosterior or a '
                'chi.HierarchicalLogPosterior')

        # Check that the log-posteriors have the same number of parameters
        self._log_posterior = log_posterior

        # Set defaults
        self._n_runs = 5
        self._parallel_evaluation = True
        self._transform = None
        self._seed = seed

        # Sample initial parameters from log-prior
        self._initial_params = self._log_posterior.sample_initial_parameters(
            n_samples=self._n_runs, seed=self._seed)

    def set_n_runs(self, n_runs):
        """
        Sets the number of times the inference routine is run.

        Each run starts from a random sample of the log-prior.
        """
        self._n_runs = int(n_runs)

        # Sample initial parameters from log-prior
        self._initial_params = self._log_posterior.sample_initial_parameters(
            n_samples=self._n_runs, seed=self._seed)

    def set_parallel_evaluation(self, run_in_parallel):
        """
        Enables or disables parallel evaluation using either a
        :class:`pints.ParallelEvaluator` or a
        :class:`pints.SequentialEvaluator`.

        If ``run_in_parallel=True``, the method will run using a number of
        worker processes equal to the detected CPU core count. The number of
        workers can be set explicitly by setting ``run_in_parallel`` to an
        integer greater than ``0``. Parallelisation can be disabled by setting
        ``run_in_parallel`` to ``0`` or ``False``.
        """
        if not isinstance(run_in_parallel, (bool, int)):
            raise ValueError(
                '`run_in_parallel` has to a boolean or an integer.')
        if run_in_parallel < 0:
            raise ValueError(
                '`run_in_parallel` cannot be negative.')

        self._parallel_evaluation = run_in_parallel

    def set_transform(self, transform):
        """
        Sets the transformation that transforms the parameter space into the
        search space.

        Transformations of the search space can significantly improve the
        performance of the inference routine.

        ``transform`` has to be an instance of `pints.Transformation` and must
        have the same dimension as the parameter space.
        """
        if not isinstance(transform, pints.Transformation):
            raise ValueError(
                'Transform has to be an instance of `pints.Transformation`.')
        if transform.n_parameters() != self._log_posterior.n_parameters():
            raise ValueError(
                'The dimensionality of the transform does not match the '
                'dimensionality of the log-posterior.')
        self._transform = transform


class OptimisationController(InferenceController):
    """
    Sets up an optimisation routine that attempts to find the parameter values
    that maximise a :class:`pints.LogPosterior`. If multiple log-posteriors are
    provided, the posteriors are assumed to be structurally identical and only
    differ due to different data sources.

    By default the optimisation is run 5 times from different initial
    starting points. Starting points are randomly sampled from the
    specified :class:`pints.LogPrior`. The optimisation is run by default in
    parallel using :class:`pints.ParallelEvaluator`.

    Extends :class:`InferenceController`.

    :param log_posterior: The log-posterior from which is sampled.
    :type log_posterior: chi.LogPosterior, chi.HierarchicalLogPosterior
    :param seed: Seed for the random initialisation. Initial points are sampled
        from the log-prior.
    :type seed: int
    """

    def __init__(self, log_posterior, seed=None):
        super(OptimisationController, self).__init__(log_posterior, seed)

        # Set default optimiser
        self._optimiser = pints.CMAES

    def run(
            self, n_max_iterations=10000, show_run_progress_bar=False,
            log_to_screen=False):
        """
        Runs the optimisation and returns the maximum a posteriori probability
        parameter estimates in from of a :class:`pandas.DataFrame` with the
        columns 'ID', 'Parameter', 'Estimate', 'Score' and 'Run'.

        The number of maximal iterations of the optimisation routine can be
        limited by setting ``n_max_iterations`` to a finite, non-negative
        integer value.

        Parameters
        ----------

        n_max_iterations
            The maximal number of optimisation iterations to find the MAP
            estimates for each log-posterior. By default the maximal number
            of iterations is set to 10000.
        show_run_progress_bar
            A boolean flag which indicates whether a progress bar for looping
            through the optimisation runs is displayed.
        log_to_screen
            A boolean flag which indicates whether the optimiser logging output
            is displayed.
        """

        # Initialise result dataframe
        result = pd.DataFrame(
            columns=['ID', 'Parameter', 'Estimate', 'Score', 'Run'])

        # Initialise intermediate container for individual runs
        run_result = pd.DataFrame(
            columns=['ID', 'Parameter', 'Estimate', 'Score', 'Run'])
        run_result['Parameter'] = self._log_posterior.get_parameter_names()

        # Set ID of individual (or IDs of parameters, if hierarchical)
        run_result['ID'] = self._log_posterior.get_id()

        # Run optimisation multiple times
        for run_id in tqdm(
                range(self._n_runs), disable=not show_run_progress_bar):
            opt = pints.OptimisationController(
                function=self._log_posterior,
                x0=self._initial_params[run_id, :],
                method=self._optimiser,
                transformation=self._transform)

            # Configure optimisation routine
            opt.set_log_to_screen(log_to_screen)
            opt.set_max_iterations(iterations=n_max_iterations)
            opt.set_parallel(self._parallel_evaluation)

            # Find optimal parameters
            try:
                estimates, score = opt.run()
            except Exception:
                # If inference breaks fill estimates with nan
                estimates = [np.nan] * self._log_posterior.n_parameters()
                score = np.nan

            # Save estimates and score of runs
            run_result['Estimate'] = estimates
            run_result['Score'] = score
            run_result['Run'] = run_id + 1
            result = pd.concat([result, run_result])

        return result

    def set_optimiser(self, optimiser):
        """
        Sets method that is used to find the maximum a posteiori probability
        estimates.
        """
        if not issubclass(optimiser, pints.Optimiser):
            raise ValueError(
                'Optimiser has to be a `pints.Optimiser`.')
        self._optimiser = optimiser


class SamplingController(InferenceController):
    """
    Sets up a sampling routine that attempts to find the posterior
    distribution of parameters defined by a :class:`pints.LogPosterior`. If
    multiple log-posteriors are provided, the posteriors are assumed to be
    structurally identical and only differ due to different data sources.

    By default the sampling is run 5 times from different initial
    starting points. Starting points are randomly sampled from the
    specified :class:`pints.LogPrior`. The optimisation is run by default in
    parallel using :class:`pints.ParallelEvaluator`.

    Extends :class:`InferenceController`.

    :param log_posterior: The log-posterior from which is sampled.
    :type log_posterior: chi.LogPosterior, chi.HierarchicalLogPosterior
    :param seed: Seed for the random initialisation. Initial points are sampled
        from the log-prior.
    :type seed: int
    """

    def __init__(self, log_posterior, seed=None):
        super(SamplingController, self).__init__(log_posterior, seed)

        # Set default sampler
        self._sampler = pints.HaarioBardenetACMC

    def _format_chains(self, chains, divergent_iters):
        """
        Formats the chains generated by pints in shape of
        (n_chains, n_iterations, n_parameters) to a xarray.Dataset, where
        each parameter name gets an entry yielding a xarray.DataArray of
        shape (n_chains, n_iterations, n_ids) or if ID is None
        (n_chains, n_iterations).

        Note that the naming of the dimensions matter for ArviZ so we call
        the dimensions (chain, draw, individual) (the last dimension
        is not set by ArviZ).
        """
        # Get all parameter names
        names = self._log_posterior.get_parameter_names()

        # Get top-level parameter names
        top_parameters = self._log_posterior.get_parameter_names(
            exclude_bottom_level=True)

        # Get IDs of bottom-level parameters
        ids = self._log_posterior.get_id(unique=True)

        # Convert names and ids to numpy arrays
        ids = np.array(ids) if isinstance(ids, list) else ids
        names = np.array(names)

        # Get the coordinates for the chains and draws
        n_chains, n_draws, _ = chains.shape
        chain_coords = list(range(n_chains))
        draw_coords = list(range(n_draws))

        # Sort samples of parameters into xarrays
        # Start with top-level parameters (have no ID)
        container = {}
        bottom_parameters = []
        for idp, parameter in enumerate(names):
            # Check if parameter is a bottom-level parameter
            if parameter not in top_parameters:
                # Add to bottom parameter list if parameter is not already in
                # it
                if parameter not in bottom_parameters:
                    bottom_parameters.append(parameter)
                continue

            # Append top parameter samples to container
            container[parameter] = xr.DataArray(
                data=chains[:, :, idp],
                dims=['chain', 'draw'],
                coords={'chain': chain_coords, 'draw': draw_coords})

        # Add samples of bottom parameters (include IDs now)
        for idp, parameter in enumerate(bottom_parameters):
            mask = names == parameter
            container[parameter] = xr.DataArray(
                data=chains[:, :, mask],
                dims=['chain', 'draw', 'individual'],
                coords={
                    'chain': chain_coords,
                    'draw': draw_coords,
                    'individual': ids})

        # Add information about divergent iterations
        attrs = {'divergent iterations': 'false'}
        if divergent_iters:
            attrs['divergent iterations'] = 'true'
            for idx, iters in enumerate(divergent_iters):
                attrs['divergent iterations chain %d' % idx] = iters

        return xr.Dataset(container, attrs=attrs)

    def run(
            self, n_iterations=10000, hyperparameters=None,
            log_to_screen=False):
        """
        Runs the sampling routine and returns the sampled parameter values in
        form of a :class:`xarray.Dataset` with :class:`xarray.DataArray`
        instances for each parameter.

        If multiple posteriors are inferred a list of :class:`xarray.Dataset`
        instances is returned.

        The number of iterations of the sampling routine can be set by setting
        ``n_iterations`` to a finite, non-negative integer value. By default
        the routines run for 10000 iterations.

        :param n_iterations: A non-negative integer number which sets the
            number of iterations of the MCMC runs.
        :type n_iterations: int, optional
        :param hyperparameters: A list of hyperparameters for the sampling
            method. If ``None`` the default hyperparameters are set.
        :type hyperparameters: list[float], optional
        :param log_to_screen: A boolean flag which can be used to print
            the progress of the runs to the screen. The progress is printed
            every 500 iterations.
        :type log_to_screen: bool, optional
        """
        # Set up sampler
        sampler = pints.MCMCController(
            log_pdf=self._log_posterior,
            chains=self._n_runs,
            x0=self._initial_params,
            method=self._sampler,
            transformation=self._transform)

        # Configure sampling routine
        sampler.set_log_to_screen(log_to_screen)
        sampler.set_log_interval(iters=20, warm_up=3)
        sampler.set_max_iterations(iterations=n_iterations)
        sampler.set_parallel(self._parallel_evaluation)

        if hyperparameters:
            for s in sampler.samplers():
                s.set_hyper_parameters(hyperparameters)

        # Run sampling routine
        chains = sampler.run()

        # If Hamiltonian Monte Carlo, get number of divergent
        # iterations
        divergent_iters = None
        if issubclass(
                self._sampler, (pints.HamiltonianMCMC, pints.NoUTurnMCMC)):
            divergent_iters = [
                s.divergent_iterations() for s in sampler.samplers()]

        # Format chains
        chains = self._format_chains(chains, divergent_iters)

        return chains

    def set_sampler(self, sampler):
        """
        Sets method that is used to sample from the log-posterior.
        """
        if not issubclass(sampler, pints.MCMCSampler):
            raise ValueError(
                'Sampler has to be a `pints.MCMCSampler`.'
            )
        self._sampler = sampler



#
# This file is part of the chi repository
# (https://github.com/DavAug/chi/) which is released under the
# BSD 3-clause license. See accompanying LICENSE.md for copyright notice and
# full license details.
#

import copy

import pandas as pd
import plotly.colors
import plotly.graph_objects as go

from chi import plots


class ParameterEstimatePlot(plots.MultiFigure):
    """
    A figure class that visualises parameter maximum a posteriori probability
    estimates across multiple optimisation runs.

    One figure is generated for each parameter, which contains a box plot of
    the parameter estimates across multiple optmisation runs. The estimates for
    each indiviudal are plotted next to each other.

    This figure can be used to assess the stability of the optimisation
    results, as well as the variation of parameter estimates across
    individuals.

    Extends :class:`MultiFigure`.
    """

    def __init__(self):
        super(ParameterEstimatePlot, self).__init__()

    def _add_box_plots(self, fig_id, data, colors):
        """
        Adds box plots of the parameter estimates for each individual across
        runs to the figure.

        One figure only contains the estimates of one parameter.
        """
        # Get number of colours in scheme
        n_colors = len(colors)

        # Add trace for each individual
        ids = data[self._id_key].unique()
        for index, individual in enumerate(ids):
            # Mask for individuals/populations parameter
            mask = data[self._id_key] == individual
            if individual is None:
                mask = data[self._id_key].isna()

            # Get data for indvidual or population parameter
            estimates = data[self._est_key][mask]
            scores = data[self._score_key][mask].to_numpy()
            runs = data[self._run_key][mask].to_numpy()
            color = colors[index % n_colors]

            self._add_trace(
                fig_id, individual, estimates, scores, runs, color)

    def _add_trace(
            self, fig_id, individual, estimates, scores, runs, color):
        """
        Adds a box plot of an individuals estimates across multiple
        optimisation runs to a figure.
        """
        # Get number of runs
        n_runs = len(runs)

        # Population parameters have an ID of None
        _id = 'Population' if individual is None else individual

        # Add trace
        fig = self._figs[fig_id]
        fig.add_trace(
            go.Box(
                y=estimates,
                name='%s' % str(_id),
                hovertemplate=(
                    'Estimate: %{y:.2f}<br>'
                    '%{text}'),
                text=[
                    'Run: %d <br>Log-posterior score: %.2f' % (
                        runs[i], scores[i]) for i in range(n_runs)],
                boxpoints='all',
                jitter=0.2,
                pointpos=-1.5,
                visible=True,
                marker=dict(
                    symbol='circle',
                    opacity=0.7,
                    line=dict(color='black', width=1)),
                marker_color=color,
                line_color=color))

    def add_data(
            self, data, id_key='ID', param_key='Parameter', est_key='Estimate',
            score_key='Score', run_key='Run'):
        """
        Adds box plots of the estimates across runs to the figure. The
        estimates are grouped by the individual ID.

        Parameters
        ----------
        data
            A :class:`pandas.DataFrame` with the parameter estimates in form of
            an ID, parameter, estimate, score, and run column.
        id_key
            Key label of the :class:`DataFrame` which specifies the ID column.
            The ID refers to the identity of an individual. Defaults to
            ``'ID'``.
        param_key
            Key label of the :class:`DataFrame` which specifies the parameter
            name column. Defaults to ``'Parameter'``.
        est_key
            Key label of the :class:`DataFrame` which specifies the parameter
            estimate column. Defaults to ``'Estimate'``.
        score_key
            Key label of the :class:`DataFrame` which specifies the score
            estimate column. The score refers to the maximum a posteriori
            probability associated with the estimate. Defaults to ``'Score'``.
        run_key
            Key label of the :class:`DataFrame` which specifies the
            optimisation run column. Defaults to ``'Run'``.
        """
        # Check input format
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                'Data has to be pandas.DataFrame.')

        keys = [param_key, id_key, est_key, score_key, run_key]
        for key in keys:
            if key not in data.keys():
                raise ValueError(
                    'Data does not have the key <' + str(key) + '>.')
        self._id_key, self._est_key, self._score_key, self._run_key = keys[1:]

        # Get a colour scheme
        colors = plotly.colors.qualitative.Plotly

        # Create one figure for each parameter
        parameters = data[param_key].unique()
        self._figs = [copy.copy(self._fig) for _ in parameters]

        # Add estimates to parameter figures
        for index, parameter in enumerate(parameters):
            # Set y label of plot to parameter name
            self._figs[index].update_layout(
                yaxis_title=parameter)

            # Get estimates for this parameter
            mask = data[param_key] == parameter
            estimates = data[mask][[id_key, est_key, score_key, run_key]]

            # Add box plots for all individuals
            self._add_box_plots(index, estimates, colors)
